var DocumentScheme = require("./DocumentScheme.js");

var Billon = {
    SoapService: require("./SoapService.js"),

    /**
     * Returns a random universally unique transfer identifier.
     * 
     * @return {string} The random universally unique identifier.
     */
    uniqueId: require('uuid/v1')
};

/**
 * Creates an instance of Billon.Payment.
 *
 * @constructor
 * @author Billon sp. z o.o.
 * @this {Billon.Payment}
 */
Billon.Payment = function() {

    /**
     * A function that run a function from the fourth argument with an object with a merchant data.
     * 
     * @param  {object} requestData An object with a request data.
     * @param  {array} items An array with products.
     * @param  {object} delivery_option A delivery data.
     * @param  {function} cb The function, which should be called with the object with a merchant data.
     */
    this.Merchant = function(requestData, items, delivery_option, cb) {
        cb({});
    };

    this.CNode = false;

    /**
     * A function that is to run after receiving response to SOAP request.
     * 
     * @param  {string} requestName A name of the SOAP message
     * @param  {object} parameters A parameters of the SOAP message
     * @param  {object} requestResponse Corporate Node response
     * @param  {object} error Errors when sending SOAP request
     */
    this.OnRequestSend = function (requestName, parameters, requestResponse, error) {};

    /**
     * A function that run a function from the second argument with the object ‘peerLocation’.
     */
    this.User = function(cb) {
        cb(false);
    };

    /**
     * A function that run a function from the second argument with an array of product ids.
     * 
     * @param {object} requestData An object with a request data.
     * @param {function} cb The function, which should be called with the array of product ids.
     */
    this.Products = function(requestData, cb) {
        cb();
    };

    /**
     * A function that run a function from the third argument with a product data or an amount expressed in an object with product data.
     */
    this.Product = function(cb) {
        cb(false);
    };

    /**
     * A function that run a function from the third argument with a product data or an amount expressed in an object with product data.
     * 
     * @param {object} requestData An object with a request data.
     * @param {(number|object)} item Product data.
     * @param {function} cb The function 
	 * 						that should be called with an object with the product data or the amount expressed in an object with product data.
     */
    this.ProductBillon = function(requestData, item, cb) {
        cb(item);
    };

    /**
     * A function that run a function from the third argument with a delivery data.
     * 
     * @param {object} requestData An object with a request data.
     * @param {array} items An array with products.
     * @param {function} cb The function 
	 * 						that should be called with an object with the delivery data.
     */
    this.ProductDelivery = function(requestData, items, cb) {
        cb(false);
    };

    /**
     * A function that run a function from the fourth argument with a path of file to download.
     * 
     * @param {object} requestData An object with a request data.
     * @param {array} items An array with products.
     * @param {object} delivery A delivery data.
     * @param {object} delivery A payment document.
     * @param {function} cb The function 
	 * 						that should be called with an object with the path of file to download.
     */
    this.ProductDownload = function(requestData, items, delivery, document, cb) {
        cb(false);
    };

    /**
     * A function that run a function from the fourth argument with the payment title.
     */
    this.PaymentTitle = function(cb) {
        cb(false);
    };

    /**
     * A function that run a function from the second argument with ISO 4217 code of the currency.
     * 
     * @param {object} requestData An object with a request data.
     * @param {function} cb The function 
	 * 						that should be called with ISO 4217 code of the currency.
     */
    this.Currency = function(requestData, cb) {
        cb('PLN');
    };
    
    /**
     * A function that run a function from the second argument with an identifier of series and the type of plate.
     * @since 2
     * 
     * @param {function} cb The function, which should be called with an identifier of series and the type of plate.
     */
    this.PlatePayment = function(requestData, cb) {
        cb('', 'NORMAL');
    };

    /**
     * A function that run a function from the third argument with false or a changed payment document.
     * 
     * @param {object} requestData An object with a request data.
     * @param {object} delivery A payment document.
     * @param {function} cb The function, which should be called with false or the changed payment document.
     */
    this.DocumentParse = function(requestData, document, cb) {
        cb(document);
    };
    
    /**
     * A function that run a function from the second argument with prices of plates.
     * @since 2
     * 
     * @param {object} requestData An object with a request data.
     * @param {function} cb The function, which should be called with the prices of the plates as arguments.
     */
    this.Plates = function(requestData, cb) {
        cb(false);
    };
    
    /**
     * A function that run a function from the second argument with the maximum amount of the plate payments in series.
     * @since 2
     * 
     * @param {object} requestData An object with a request data.
     * @param {function} cb The function that should be called with the maximum amount, 
	 * 						which is expressed in the number of thousandths of a basic monetary unit, of the plate payments in series.
     */
    this.MaxTotalAmount = function(requestData, cb) {
        cb(0);
    };
    
    /**
     * A function that run a function from the second argument with the duration of the series the plate payments in seconds.
     * @since 2
     * 
     * @param {object} requestData An object with a request data.
     * @param {function} cb The function that should be called with the duration of the series the plate payments in seconds.
     */
    this.Duration = function(requestData, cb) {
        cb(0);
    };
    
    /**
     * A function that run a function from second argument with the maximum number of payments in a series of the plate payments.
     * @since 2
     * 
     * @param {object} requestData An object with a request data.
     * @param {function} cb The function that should be called with the maximum number of payments in a series of the plate payments.
     */
    this.MaxPaymentsInSeries = function(requestData, cb) {
        cb(0);
    };
    
    /**
     * A function that run a function from the second argument with an name of a brand.
     * @since 2
     * 
     * @param {object} requestData An object with a request data.
     * @param {function} cb The function that should be called with the name of a brand.
     */
    this.MerchantBrandName = function(requestData, cb) {
        cb(false);
    };

    /**
     * A function that is to run when the payment is successfully completed.
     */
    this.OnFinishedPayment = function() {
        console.log("Finished some payment.");
    };
    
    /**
     * A function to handle the status returned by Corporate Node during the operation of acceptance of the plate payments series.
     * @since 2
     */
    this.OnReceivedPlatePaymentSeriesStatus = function() {};
    
    /**
     * A function that is to run when the plate payment is successfully accepted.
     * @since 2
     */
    this.OnAcceptedPlatePaymentSeries = function() {};

};

/**
 * Checks if variable is function.
 * 
 * @private
 * @author Billon sp. z o.o.
 * @param func The variable to check
 * @return {boolean} True if variable is function, otherwise false.
 */
var isFunction = function(func) {
    if (typeof func !== "function") {
        return false;
    }

    return true;
};

var deliveryOptions = {
    "NOT_NEEDED": 0,
    "OPTIONAL": 1,
    "OBLIGATORY": 2
};

var uids = [];

var taskTracker = {};

/**
 * Returns a name for merchant's brand from the second argument if exists or the domain of a page that has sent the request otherwise
 * @since 2
 * 
 * @private
 * @author Billon sp. z o.o.
 * @param {object} headers The Sails.js headers in request.
 * @param {(string|undefined|null|false)} mbn The name for merchant's brand.
 * @return {string} The name for merchant's brand.
 */
var getMerchantBrandName = function(headers, mbn) {
   if (mbn) {
      return mbn;
   } else {
      var url = require('url');
      return url.parse(headers['origin'] ? headers['origin'] : headers['referer']).hostname;
   }
};

/**
 * Validates a peer location and adds it to parameters of a request if peer location is correct.
 * 
 * @private
 * @author Billon sp. z o.o.
 * @param {object} user The peer location.
 * @param {object} params The parameters of a payment request or request of start of payment series.
 */
var addPeerLocationIfAvailable = function(user, params) {
    if (typeof user.confirmationCode1 != 'undefined' && user.confirmationCode1 != null && typeof user.confirmationCode2 != 'undefined' && user.confirmationCode2 != null) {
      if (!user.userLocation || !user.userLocation.machineIdHex || !user.userLocation.ip || !user.userLocation.tcp || !user.userLocation.udp) {
         user.userLocation = null;
      }
      if (!user.buddyLocation || !user.buddyLocation.machineIdHex || !user.buddyLocation.ip || !user.buddyLocation.tcp || !user.buddyLocation.udp) {
         user.buddyLocation = null;
      }
      params.peerLocation = user;
    }
};

// ERRORS
var errs = {
    ERR_INVALID_BODY: {
        err: "ERR_INVALID_BODY",
        description: "Missing body."
    },
    ERR_INVALID_QUERY: {
        err: "ERR_INVALID_QUERY",
        description: "Missing query."
    },
    ERR_INVALID_QUERY_DATA: {
        err: "ERR_INVALID_QUERY_DATA",
        description: "Data must be an object."
    },
    ERR_INVALID_PRODUCTS: {
        err: "ERR_INVALID_PRODUCTS",
        description: "Data.products must be an object."
    },
    ERR_INVALID_USER: {
        err: "ERR_INVALID_USER",
        description: "Data.user must be an object."
    },
    ERR_INVALID_WUI: {
        err: "ERR_INVALID_WUI",
        description: "Data.cNode must be an object."
    },
    ERR_INVALID_USER_USERNAME: {
        err: "ERR_INVALID_USER_USERNAME",
        description: "Data.user.userName must be a string with minimum 4 characters length."
    },
    ERR_INVALID_TASKID: {
        err: "ERR_INVALID_TASKID",
        description: "Invalid taskId sent."
    },
    ERR_INVALID_UID: {
        err: "ERR_INVALID_UID",
        description: "Invalid uid sent."
    },
    ERR_TASK_DOES_NOT_EXIST: {
        err: "ERR_TASK_DOES_NOT_EXIST",
        description: "Task does not exist."
    },
    ERR_CANNOT_DOWNLOAD: {
        err: "ERR_CANNOT_DOWNLOAD",
        description: "Cannot download file."
    },
    ERR_INVALID_AMOUNT: {
        err: "ERR_INVALID_AMOUNT",
        description: "Invalid amount."
    }
};

var plateColors = {'PLATE_ALPHA': '44B905', 'PLATE_BETA': '00ADFF', 'PLATE_GAMMA': 'B466CB', 'PLATE_DELTA': 'F6891C', 'PLATE_EPSILON': 'EC2B41'};

Billon.Payment.prototype = {
    
    /**
     * Sets a function that run a function from the fourth argument with an object with a merchant data.
     * 
     * @param {function} func The function with four arguments: 
	 * 						  an object with a request data, an array with products, 
	 * 						  a delivery data and the function, which should be called with the object with a merchant data.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    merchant: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.merchant parameter must be a function.");
            return false;
        }

        this.Merchant = func;
        return this;
    },

    /**
     * Sets a function that run a function from the fourth argument with API key of the user of the service ‘billon.me’ 
	 * or an object with username with which the application 24/7 is signed in and address on which the application 24/7 listens.
     * 
     * @param {function} func The function with four arguments:
	 * 						  an object with a request data, an array with products, 
	 * 						  a delivery data and the function, which should be called with API key of the user of the service ‘billon.me’ 
	 *  					  or an object with username with which the application 24/7 is signed in and address on which the application 24/7 listens.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    cNode: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.cNode parameter must be a function.");
            return false;
        }

        this.CNode = func;
        return this;
    },

    /**
     * Sets a function that run a function from the fourth argument with API key of the user of the service ‘billon.me’ 
	 * or an object with username with which the application 24/7 is signed in and address on which the application 24/7 listens.
     * @deprecated since version 2.5
     * 
     * @param {function} func The function with four arguments:
	 * 						  an object with a request data, an array with products, 
	 * 						  a delivery data and the function, which should be called with API key of the user of the service ‘billon.me’ 
	 *  					  or an object with username with which the application 24/7 is signed in and address on which the application 24/7 listens.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    app24: function(func) {
        return this.cNode(func);
    },

    /**
     * Sets a function that is to run after receiving response to SOAP request.
     * @since 2.1
     * 
     * @param {function} func The function with four arguments:
	 * 						  name of the SOAP message, parameters of the SOAP message, Corporate Node response, errors when sending SOAP request.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    onRequestSend: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.onRequestSend parameter must be a function.");
            return this;
        }

        this.OnRequestSend = func;
        return this;
    },

    /**
     * Sets a function that run a function from the second argument with the object ‘peerLocation’.
     * 
     * @param {function} func The function with two arguments: 
	 * 						  an object with a request data and the function, which should be called with the object ‘peerLocation’.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    user: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.user parameter must be a function.");
            return false;
        }

        this.User = func;
        return this;
    },
    
    /**
     * Sets a function that run a function from the second argument with an array of product ids.
     * 
     * @param {function} func The function with two arguments: 
	 * 						  an object with a request data and the function, which should be called with the array of product ids.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    products: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.products parameter must be a function.");
            return false;
        }

        this.Products = func;
        return this;
    },

    /**
     * Sets a function that run a function from the third argument with a product data or an amount expressed in an object with product data.
     * 
     * @param {function} func The function with two arguments: 
	 * 						  an object with a request data 
	 * 						  and the function, 
	 * 						  which should be called with an object with the product data or the amount expressed in an object with product data.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    product: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.product parameter must be a function.");
            return false;
        }

        this.Product = func;
        return this;
    },

    /**
     * Sets a function that run a function from the third argument with a product data or an amount expressed in an object with product data.
     * 
     * @param {function} func The function with three arguments: 
	 * 						  an object with a request data, a product data
	 * 						  and the function, 
	 * 						  which should be called with an object with the product data or the amount expressed in an object with product data.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    productBillon: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.productBillon parameter must be a function.");
            return false;
        }

        this.ProductBillon = func;
        return this;
    },

    /**
     * Sets a function that run a function from the third argument with a delivery data.
     * 
     * @param {function} func The function with three arguments: 
	 * 						  an object with a request data, an object with product data 
	 * 						  and the function, which should be called with an object with the delivery data.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    productDelivery: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.productDelivery parameter must be a function.");
            return false;
        }

        this.ProductDelivery = func;
        return this;
    },

    /**
     * Sets a function that run a function from the fourth argument with an array with a path of file to download.
     * 
     * @param {function} func The function with four arguments: 
	 * 						  an object with a request data, an object with product data, a payment document, 
	 * 						  and the function, which should be called with an object with the path of file to download.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    productDownload: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.productDownload parameter must be a function.");
            return false;
        }

        this.ProductDownload = func;
        return this;
    },

    /**
     * Sets a function that run a function from the second argument with ISO 4217 code of the currency.
     * 
     * @param {function} func The function with two arguments: 
	 *                        an object with a request data and the function, which should be called with ISO 4217 code of the currency.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    currency: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.currency parameter must be a function.");
            return this;
        }

        this.Currency = func;
        return this;
    },

	/**
     * Sets a function that run a function from the second argument with an identifier of series and the type of plate.
     * @since 2
     * 
     * @param {function} func The function with two arguments: 
	 *                        an object with a request data and the function, which should be called with an identifier of series and the type of plate.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    platePayment: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.platePayment parameter must be a function.");
            return this;
        }

        this.PlatePayment = func;
        return this;
    },

    /**
     * Sets a function that run a function from the third argument with false or a changed payment document.
     * 
     * @param {function} func The function with three arguments: 
	 *                        an object with a request data, the payment document 
	 * 						  and the function, which should be called with false or the changed payment document.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    documentParse: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.documentParse parameter must be a function.");
            return false;
        }

        this.DocumentParse = func;
        return this;
    },

    /**
     * Sets a function that run a function from the fourth argument with the payment title.
     * 
     * @param {function} func The function with four arguments: 
	 * 						  an object with a request data, an array with products, 
	 * 						  a delivery data and the function, which should be called with the payment title.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    paymentTitle: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.paymentTitle parameter must be a function.");
            return false;
        }

        this.PaymentTitle = func;
        return this;
    },

    /**
     * Sets a function that run a function from the second argument with prices of plates.
     * @since 2
     * 
     * @param {function} func The function with two arguments: 
	 *                        an object with a request data and the function, which should be called with the prices of the plates as arguments.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    plates: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.plates parameter must be a function.");
            return this;
        }

        this.Plates = func;
        return this;
    },

    /**
     * Sets a function that run a function from the second argument with the maximum amount of the plate payments in series.
     * 
     * @param {function} func The function with two arguments: 
	 *                        an object with a request data 
	 * 						  and the function, which should be called with the maximum amount, 
	 * 						  which is expressed in the number of thousandths of a basic monetary unit, of the plate payments in series.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    maxTotalAmount: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.maxTotalAmount parameter must be a function.");
            return this;
        }

        this.MaxTotalAmount = func;
        return this;
    },

    /**
     * Sets a function that run a function from the second argument with the duration of the series the plate payments in seconds.
     * @since 2
     * 
     * @param {function} func The function with two arguments: 
	 *                        an object with a request data 
	 * 						  and the function, which should be called with the duration of the series the plate payments in seconds.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    duration: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.duration parameter must be a function.");
            return this;
        }

        this.Duration = func;
        return this;
    },

    /**
     * Sets a function that run a function from second argument with the maximum number of payments in a series of the plate payments.
     * @since 2
     * 
     * @param {function} func The function with two arguments: 
	 *                        an object with a request data 
	 * 						  and the function, which should be called with the maximum number of payments in a series of the plate payments.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    maxPaymentsInSeries: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.maxPaymentsInSeries parameter must be a function.");
            return this;
        }

        this.MaxPaymentsInSeries = func;
        return this;
    },

    /**
     * Sets a function that run a function from the second argument with an name of a brand.
     * @since 2
     * 
     * @param {function} func The function with two arguments: 
	 * 						  an object with a request data and the function, which should be called with the name of a brand.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    merchantBrandName: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.merchantBrandName parameter must be a function.");
            return this;
        }

        this.MerchantBrandName = func;
        return this;
    },

    /**
     * Sets a function that is to run when the payment is successfully completed.
     * 
     * @param {function} func The function with four arguments: products, a delevery data, a purchase docuemnt and a request data.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    onFinishedPayment: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.onFinishedPayment parameter must be a function.");
            return false;
        }

        this.OnFinishedPayment = func;
        return this;
    },

    /**
     * Sets a function to handle the status returned by Corporate Node during the operation of acceptance of the plate payments series.
     * @since 2
     * 
     * @param {function} func The function with an object with the response as argument.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    onReceivedPlatePaymentSeriesStatus: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.onReceivedPlatePaymentSeriesStatus parameter must be a function.");
            return false;
        }

        this.OnReceivedPlatePaymentSeriesStatus = func;
        return this;
    },

    /**
     * Sets a function that is to run when the plate payment is successfully accepted.
     * @since 2
     * 
     * @param {function} func The function with an object with the payment data as argument.
     * @return {(Billon.Payment|false)} The object of this method.
     */
    onAcceptedPlatePaymentSeries: function(func) {
        if (!isFunction(func)) {
            console.log("Payment.onAcceptedPlatePaymentSeries parameter must be a function.");
            return false;
        }

        this.OnAcceptedPlatePaymentSeries = func;
        return this;
    },

    /**
     * Initiates and listen to donation for user's object and value sended by POST.
     * 
	 * @param {Request} req The first parameter of a function of Sails.js controller, Sails.js incoming request object.
	 * @param {Response} res The second parameter of a function of Sails.js controller, Sails.js outgoing response object.
     */
    donateInitiation: function(req, res) {
        var self = this;

        if (typeof req.body !== "object") {
            return res.send(errs.ERR_INVALID_QUERY);
        }

        var requestData = req.body;

        if (typeof requestData.user !== "object") {
            return res.send(errs.ERR_INVALID_USER);
        }

        if (typeof requestData.amount !== "number") {
            return res.send(errs.ERR_INVALID_AMOUNT);
        }

        if (typeof requestData.user.userName !== "string" || requestData.user.userName.length < 4) {
            return res.send(errs.ERR_INVALID_USER_USERNAME);
        }

        var paymentData = {
            user: requestData.user,
            amount: requestData.amount
        };

        var requestInvoiceData = "NOT_NEEDED";
        var requestDeliveryData = "NOT_NEEDED";
        self.CNode(requestData, null, null, function(wui) {
            var document = DocumentScheme();
            self.DocumentParse(requestData, document, function(parsedDocument) {
                if (parsedDocument !== false) {
                    document = parsedDocument;
                }

                // Wczytanie danych sprzedawcy.
                self.Merchant(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(merchant) {
                    var uid = Billon.uniqueId();
                    document.sales_document_data.seller_data = merchant;
                    document.sales_document_data.sales_document_id = uid;

                    // Ustalenie daty.
                    document.sales_document_data.sales_document_date = new Date().getTime();
                    document.sales_document_data.trade_date = new Date().getTime();
                    document.sales_document_data.total_amount_gross = paymentData.amount;

                    self.PaymentTitle(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(title) {
                        self.Currency(requestData, function(currency) {
                           self.MerchantBrandName(requestData, function(mbn) {
                            var data = {
                                cNode: wui,
                                params: {
                                    type: "NORMAL",
                                    amount: {
                                        amount: paymentData.amount,
                                        currency: currency
                                    },
                                    requestData: {
                                        username: paymentData.user.userName,
                                        uniqueTransferId: document.sales_document_data.sales_document_id,
                                        title: title,
                                        merchantBrandName: getMerchantBrandName(req.headers, mbn),
                                        requestInvoiceData: requestInvoiceData,
                                        requestDeliveryData: requestDeliveryData,
                                        description: document ? JSON.stringify(document) : null
                                    }
                                }
                            };
                            addPeerLocationIfAvailable(paymentData.user, data.params);
                            Billon.SoapService.onRequestSend = self.OnRequestSend;
                            Billon.SoapService.call.requestPaymentInitiation(data, function(requestResponse) {
                                var copy = _.clone(requestResponse);
                                if (requestResponse.requestStatus === "SUCCESS") {
                                    taskTracker[uid] = {
                                        taskId: requestResponse.taskId,
                                        cNode: wui,
                                        paymentData: paymentData
                                    };
                                    var interval = setInterval(function() {
                                        Billon.SoapService.onRequestSend = self.OnRequestSend;
                                        Billon.SoapService.call.getTaskStatus({
                                            cNode: wui,
                                            params: {
                                                taskId: requestResponse.taskId
                                            }
                                        }, function(taskResponse) {
                                            if (taskResponse.requestStatus !== "SUCCESS") {
                                                setTimeout(function() {
                                                    delete taskTracker[uid];
                                                }, 30000);
                                                return clearInterval(interval);
                                            }

                                            if (taskResponse.status.split("_")[0] === "FINISHED" || taskResponse.progressPercent === 100 || taskResponse.progressPercent === "100") {
                                                clearInterval(interval);
                                                setTimeout(function() {
                                                    delete taskTracker[uid];
                                                }, 30000);
                                            }

                                            if (taskResponse.status === "FINISHED_OK") {
                                                document.sales_document_data.items = paymentData;
                                                document.additionalInfo = taskResponse.additionalInfo;
                                                document.taskId = requestResponse.taskId;
                                                self.OnFinishedPayment(document.sales_document_data.items, document.sales_document_data.delivery_option, document, requestData);
                                            }

                                        });
                                    }, 5000);
                                    copy.taskId = uid;
                                }

                                res.json(copy);
                            });
                           });
                        });
                    });
                });
            });
        });
    },

    /**
     * Initiates and listen to payment.
     * 
	 * @param {Request} req The first parameter of a function of Sails.js controller, Sails.js incoming request object.
	 * @param {Response} res The second parameter of a function of Sails.js controller, Sails.js outgoing response object.
     */
    paymentInitiation: function(req, res) {
        var self = this;
        if (typeof req.body !== "object") {
            return res.send(errs.ERR_INVALID_BODY);
        }

        var requestData = req.body,
            paymentData = {};

        self.User(requestData, function(user) {
            if (typeof user !== "object") {
                return res.send(errs.ERR_INVALID_USER);
            }

            if (typeof user.userName !== "string" || user.userName.length < 4) {
                return res.send(errs.ERR_INVALID_USER_USERNAME);
            }

            paymentData.user = user;

            self.Products(requestData, function(products) {
                if (products && typeof products !== "object") {
                    return res.send(errs.ERR_INVALID_PRODUCTS);
                }

                paymentData.products = products;

                var document = DocumentScheme();

                self.DocumentParse(requestData, document, function(parsedDocument) {
                    if (parsedDocument !== false) {
                        document = parsedDocument;
                    }

                    var requestInvoiceData = "NOT_NEEDED";
                    var requestDeliveryData = "NOT_NEEDED";

                    // Wczytanie produktów.
                    var itemsLoaded = 0;
                    paymentData.items = [];
                    var productsLength = 1;
                    var onItemReceived = function(item) {
                        paymentData.items.push(item);
                        self.ProductBillon(requestData, item, function(currentItem) {
                            if (typeof currentItem == "object") {
                                currentItem.unit_price_net = parseInt(currentItem.unit_price_net * 1000);
                                currentItem.amount_net = currentItem.unit_price_net * currentItem.quantity;
                                document.sales_document_data.total_amount_net += currentItem.amount_net;
                                currentItem.amount_gross = currentItem.amount_gross * 1000;
                                
                                // Sprawdzenie czy potrzeba danych do faktury lub dostawy.
                                if (deliveryOptions[currentItem.requestInvoiceData] > deliveryOptions[requestInvoiceData]) {
                                    requestInvoiceData = currentItem.requestInvoiceData;
                                }

                                if (deliveryOptions[currentItem.requestDeliveryData] > deliveryOptions[requestDeliveryData]) {
                                    requestDeliveryData = currentItem.requestDeliveryData;
                                }

                                if (currentItem.vat && currentItem.vat.amount >= 0) {
                                    currentItem.vat.amount *= 1000;
                                }
                                // Dodanie produktu do listy.
                                document.sales_document_data.items.push(currentItem);
                                document.sales_document_data.total_amount_gross += currentItem.amount_gross * 
                                                                                   (currentItem.quantity || currentItem.quantity === 0 ? 
                                                                                    currentItem.quantity : 1);
                            }
                            else {
                                document.sales_document_data.items.push({amount_gross: currentItem});
                                document.sales_document_data.total_amount_gross += currentItem;
                            }

                            itemsLoaded++;
                            if (itemsLoaded === productsLength) {
                                onItemsLoaded();
                            }
                        });
                    };
                    if (paymentData.products) {
                        productsLength = paymentData.products.length;
                        for (var i = 0; i < productsLength; i++) {
                            var itemId = paymentData.products[i];
                            self.Product(requestData, itemId, onItemReceived);
                        }
                    }
                    else {
                        self.Product(requestData, undefined, onItemReceived);
                    }
                    function onItemsLoaded() {

                        // Wczytanie opcji dostawy.
                        self.ProductDelivery(requestData, document.sales_document_data.items, function(delivery) {
                            document.sales_document_data.delivery_option = delivery;

                            // Wczytanie danych sprzedawcy.
                            self.Merchant(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(merchant) {
                                var uid = Billon.uniqueId();
                                document.sales_document_data.seller_data = merchant;
                                document.sales_document_data.sales_document_id = uid;

                                // Ustalenie daty.
                                document.sales_document_data.sales_document_date = new Date().getTime();
                                document.sales_document_data.trade_date = new Date().getTime();

                                self.CNode(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(wui) {
                                    if (!wui) {
                                        return res.send(errs.ERR_INVALID_WUI);
                                    }
                                    self.PaymentTitle(requestData, document.sales_document_data.items, document.sales_document_data.delivery_option, function(title) {
                                        self.Currency(requestData, function(c) {
                                          self.PlatePayment(requestData, function(spid, t) {
                                             self.MerchantBrandName(requestData, function(mbn) {
                                               var keyCodeArr = [];
                                               var counter = 0;
                                               var keyCodeItems = _.filter(paymentData.items, function(item) {
                                                   return item.KeyCodes;
                                               });
                                               var keyCodeItemsLength = keyCodeItems.length;

                                               var errorCalled = false;
                                               var onKeyCodeDownload = function(download) {
                                                   if (errorCalled) {
                                                       return;
                                                   }

                                                   if (download.err) {
                                                       errorCalled = true;
                                                       return res.json(errs.ERR_CANNOT_DOWNLOAD);
                                                   }

                                                   if (!download || download.length === 0) {
                                                       return res.json(errs.ERR_CANNOT_DOWNLOAD);
                                                   }

                                                   // TODO: Obsłużyć wiele kodów jako 1 produkt.
                                                   KeyCode.update({
                                                       id: download[0].id
                                                   }, {
                                                       Status: "blocked"
                                                   }).exec(function(err, updated) {
                                                       if (err) {
                                                           console.log(err);
                                                           return;
                                                       }

                                                       if (typeof updated !== "undefined") {
                                                           keyCodeArr.push(updated[0]);
                                                       }

                                                       counter++;
                                                       if (keyCodeItemsLength === counter) {
                                                           onDownloadedFinished();
                                                       }
                                                   });
                                               };

                                               if (keyCodeItemsLength === 0) {
                                                   return self.ProductDownload(requestData, paymentData.items, delivery, document, onDownloadedFinished);
                                               }

                                               for (var i = 0; i < keyCodeItemsLength; i++) {
                                                   self.ProductDownload(requestData, paymentData.items, delivery, document, onKeyCodeDownload);
                                               }

                                               function onDownloadedFinished(downloadArr) {
                                                   if (t) {
                                                      t = t.toUpperCase();
                                                      if (t == 'ALPHA' || t == 'BETA' || t == 'GAMMA' || t == 'DELTA' || t == 'EPSILON') {
                                                         t = 'PLATE_' + t;
                                                      }
                                                   }
                                                   var data = {
                                                       cNode: wui,
                                                       params: {
                                                           type: t ? t : "NORMAL",
                                                           amount: {
                                                               amount: document.sales_document_data.total_amount_gross,
                                                               currency: c
                                                           },
                                                           requestData: {
                                                               username: paymentData.user.userName,
                                                               uniqueTransferId: uid,
                                                               title: title,
                                                               merchantBrandName: getMerchantBrandName(req.headers, mbn),
                                                               requestInvoiceData: requestInvoiceData,
                                                               requestDeliveryData: requestDeliveryData,
                                                               description: JSON.stringify(document)
                                                           }
                                                       }
                                                   };
                                                   if (t && t.indexOf('PLATE_') == 0) {
                                                      data.params.requestData.plateColorHex = plateColors[t];
                                                   }
                                                   if (spid) {
                                                      data.params.seriesId = spid;
                                                   }
                                                   addPeerLocationIfAvailable(paymentData.user, data.params);
                                                   Billon.SoapService.onRequestSend = self.OnRequestSend;
                                                   Billon.SoapService.call.requestPaymentInitiation(data, function(requestResponse) {
                                                       var onUpdated = function(err, updated) {
                                                           if (err) {
                                                               console.log(err);
                                                           }
                                                       };

                                                       var unBlockKeyCodes = function() {
                                                           for (var i = 0; i < keyCodeArr.length; i++) {
                                                               var keyCode = keyCodeArr[i];
                                                               KeyCode.update({
                                                                   id: keyCode.id
                                                               }, {
                                                                   Status: "available"
                                                               }).exec(onUpdated);
                                                           }
                                                       };

                                                       if (requestResponse.err || requestResponse.requestStatus !== "SUCCESS") {
                                                           unBlockKeyCodes();
                                                       }

                                                       var copy = _.clone(requestResponse);
                                                       if (requestResponse.requestStatus === "SUCCESS") {
                                                           taskTracker[uid] = {
                                                               requestData: requestData,
                                                               taskId: requestResponse.taskId,
                                                               cNode: wui,
                                                               items: paymentData.items,
                                                               delivery: document.sales_document_data.delivery_option,
                                                               document: document,
                                                               download: downloadArr,
                                                               keyCodes: keyCodeArr
                                                           };

                                                           var interval = setInterval(function() {
                                                               Billon.SoapService.onRequestSend = self.OnRequestSend;
                                                               Billon.SoapService.call.getTaskStatus({
                                                                   cNode: wui,
                                                                   params: {
                                                                       taskId: requestResponse.taskId
                                                                   }
                                                               }, function(taskResponse) {
                                                                   if (taskResponse.requestStatus !== "SUCCESS") {
                                                                       clearInterval(interval);
                                                                       setTimeout(function() {
                                                                           delete taskTracker[uid];
                                                                       }, 30000);

                                                                       unBlockKeyCodes();
                                                                       return;
                                                                   }

                                                                   var onUpdatedKey = function(err, updated) {
                                                                       if (err) {
                                                                           console.log(err);
                                                                       }
                                                                   };

                                                                   var keyCodeUsed;

                                                                   if (taskResponse.status.split("_")[0] === "FINISHED" || taskResponse.progressPercent === 100 
                                                                       || taskResponse.progressPercent === "100") {
                                                                       clearInterval(interval);
                                                                       setTimeout(function() {
                                                                           delete taskTracker[uid];
                                                                       }, 30000);

                                                                       if (taskResponse.status !== "FINISHED_OK") {
                                                                           for (var j = 0; j < keyCodeArr.length; j++) {
                                                                               keyCodeUsed = keyCodeArr[j];
                                                                               KeyCode.update({
                                                                                   id: keyCodeUsed.id
                                                                               }, {
                                                                                   Status: "available"
                                                                               }).exec(onUpdatedKey);
                                                                           }
                                                                       }

                                                                   }

                                                                   if (taskResponse.status === "FINISHED_OK") {
                                                                       for (var k = 0; k < keyCodeArr.length; k++) {
                                                                           keyCodeUsed = keyCodeArr[k];
                                                                           console.log("keyCodeUsed");
                                                                           console.log(keyCodeUsed);
                                                                           KeyCode.update({
                                                                               id: keyCodeUsed.id
                                                                           }, {
                                                                               Status: "sold"
                                                                           }).exec(onUpdatedKey);
                                                                       }

                                                                       document.additionalInfo = taskResponse.additionalInfo;
                                                                       document.taskId = requestResponse.taskId;
                                                                       self.OnFinishedPayment(document.sales_document_data.items, document.sales_document_data.delivery_option, document, requestData);
                                                                   }

                                                               });
                                                           }, 5000);
                                                           copy.taskId = uid;
                                                       }

                                                       res.json(copy);

                                                   });
                                               }
                                             });
                                          });
                                        });
                                    });
                                });
                            });
                        });
                    }
                });
            });
        });
    },

    /**
     * Outputs information about a tariff of plate payments.
     * @since 2.1
     * 
	 * @param {Response} req The second parameter of a function of Sails.js controller, Sails.js outgoing response object.
	 * @param {Response} res The second parameter of a function of Sails.js controller, Sails.js outgoing response object.
     */
    tariff: function(req, res) {
        var self = this;
        if (self.tariff.arguments.length == 1) {
            res = req;
        }
        self.Plates({}, function(alpha, beta, gamma, delta, epsilon) {
            self.Currency({}, function(c) {
                self.Duration({}, function(d) {
                   self.MaxTotalAmount({}, function(mta) {
                       self.MerchantBrandName({}, function(mbn) {
                            res.json({amounts: [alpha, beta, gamma, delta, epsilon], currency: c, duration: d, maxTotalAmount: mta, merchantBrandName: mbn});
                       });
                   });
                });
            });
        });
    },

    /**
     * Outputs information about a tariff of plate payments.
     * @deprecated since version 2.1
     * @since 2
     * 
	 *
	 * @param {Response} req The second parameter of a function of Sails.js controller, Sails.js outgoing response object.
	 * @param {Response} res The second parameter of a function of Sails.js controller, Sails.js outgoing response object.
     */
    platesList: function(req, res) {
        if (this.platesList.arguments.length == 1) {
            this.tariff(req, req);
        }
        else {
            this.tariff(null, res);
        }
    },

    /**
     * Initiates and listen to start of series of plate payments.
     * @since 2
     * 
	 * @param {Request} req The first parameter of a function of Sails.js controller, Sails.js incoming request object.
	 * @param {Response} res The second parameter of a function of Sails.js controller, Sails.js outgoing response object.
     */
    platePaymentSeriesStart: function(req, res) {
        var self = this;
        if (typeof req.body !== "object") {
            return res.send(errs.ERR_INVALID_BODY);
        }
        var requestData = req.body,
            paymentData = {};
        self.User(requestData, function(user) {
            if (typeof user !== "object") {
                return res.send(errs.ERR_INVALID_USER);
            }

            if (typeof user.userName !== "string" || user.userName.length < 4) {
                return res.send(errs.ERR_INVALID_USER_USERNAME);
            }
            paymentData.user = user;
            self.Plates(requestData, function(alpha, beta, gamma, delta, epsilon) {
               var cpc = _.clone(plateColors);
               if (!beta) {
                  beta = alpha;
                  cpc.PLATE_BETA = cpc.PLATE_ALPHA;
                  alpha = 0;
               }
               if (!gamma) {
                  gamma = beta;
                  beta = alpha;
                  cpc.PLATE_GAMMA = cpc.PLATE_BETA;
                  cpc.PLATE_BETA = cpc.PLATE_ALPHA;
                  alpha = 0;
               }
               if (!delta) {
                  delta = gamma;
                  gamma = beta;
                  beta = alpha;
                  cpc.PLATE_DELTA = cpc.PLATE_GAMMA;
                  cpc.PLATE_GAMMA = cpc.PLATE_BETA;
                  cpc.PLATE_BETA = cpc.PLATE_ALPHA;
                  alpha = 0;
               }
               if (!epsilon) {
                  epsilon = delta;
                  delta = gamma;
                  gamma = beta;
                  beta = alpha;
                  cpc.PLATE_EPSILON = cpc.PLATE_DELTA;
                  cpc.PLATE_DELTA = cpc.PLATE_GAMMA;
                  cpc.PLATE_GAMMA = cpc.PLATE_BETA;
                  cpc.PLATE_BETA = cpc.PLATE_ALPHA;
                  alpha = 0;
               }
               var plates = {'alpha-price': alpha, 'beta-price': beta, 'gamma-price': gamma, 'delta-price': delta, 'epsilon-price': epsilon,
                             'alpha-color': cpc.PLATE_ALPHA, 'beta-color': cpc.PLATE_BETA, 'gamma-color': cpc.PLATE_GAMMA, 
                             'delta-color': cpc.PLATE_DELTA, 'epsilon-color': cpc.PLATE_EPSILON};
               self.CNode(requestData, plates, null, function(wui) {
                        if (!wui) {
                           return res.send(errs.ERR_INVALID_WUI);
                        }
                        self.Currency(requestData, function(c) {
                           self.MaxTotalAmount(requestData, function(mta) {
                              self.Duration(requestData, function(d) {
                                 self.MaxPaymentsInSeries(requestData, function(mp) {
                                    self.MerchantBrandName(requestData, function(t) {
                                       var title = getMerchantBrandName(req.headers, t);
                                       var uid = Billon.uniqueId();
                                       data = {cNode: wui, 
                                               params: {seriesData: {
                                                            description: ' ', 
                                                            username: paymentData.user.userName, 
                                                            maxTotalAmount: {amount: mta, currency: c, colour: 0}, 
                                                            maxPaymentsInSeries: mp,
                                                            seriesSecondDuration: d, 
                                                            plateParameters: plates}, 
                                                        type: 'PLATE_PAYMENT_SERIES', 
                                                        requestData: {
                                                            username: paymentData.user.userName,
                                                            uniqueTransferId: uid,
                                                            title: title,
                                                            merchantBrandName: title,
                                                            requestInvoiceData: 'NOT_NEEDED',
                                                            requestDeliveryData: 'NOT_NEEDED',
                                                            description: ' ', 
                                                            channelInfo: ' ', 
                                                            antiSpamCode: ' '}}};
                                       addPeerLocationIfAvailable(paymentData.user, data.params);
                                       Billon.SoapService.onRequestSend = self.OnRequestSend;
                                       Billon.SoapService.call.startSeriesPayment(data, function(requestResponse) {
                                            var copy = _.clone(requestResponse);
                                            if (requestResponse.requestStatus === "SUCCESS") {
                                                taskTracker[uid] = {
                                                    taskId: requestResponse.taskId,
                                                    cNode: wui,
                                                    paymentData: paymentData
                                                };
                                                var interval = setInterval(function() {
                                                    Billon.SoapService.onRequestSend = self.OnRequestSend;
                                                    Billon.SoapService.call.getTaskStatus({
                                                        cNode: wui,
                                                        params: {
                                                            taskId: requestResponse.taskId
                                                        }
                                                    }, function(taskResponse) {
                                                        self.OnReceivedPlatePaymentSeriesStatus(taskResponse);
                                                        if (taskResponse.requestStatus !== "SUCCESS") {
                                                            setTimeout(function() {
                                                                delete taskTracker[uid];
                                                            }, 30000);
                                                            return clearInterval(interval);
                                                        }
                                                        var PaymentAccepted = taskResponse.status === "TRANSFERRING" || taskResponse.status === "TRANSFERING"
                                                                                                      || taskResponse.status === "SERIES_ACCEPTED";
                                                        if (PaymentAccepted || taskResponse.status.split("_")[0] === "FINISHED" || 
                                                            taskResponse.progressPercent === 100 || taskResponse.progressPercent === "100") {
                                                            clearInterval(interval);
                                                            setTimeout(function() {
                                                                delete taskTracker[uid];
                                                            }, 30000);
                                                        }
                                                        if (PaymentAccepted) {
                                                            self.OnAcceptedPlatePaymentSeries(paymentData);
                                                        }

                                                    });
                                                }, 5000);
                                                copy.taskId = uid;
                                            }
                                            res.json(copy);
                                       });
                                 });
                              });
                           });
                        });
               });
            });
         });
      });
    },

    /**
     * Checks status of a payment or a start of plate payment series.
     * 
	 * @param {string} $uid An identifier of the task.
	 * @param {function} $cb A function, which is to be called with a response of Corporate Node or a service 'billon.me'.
     */
    paymentStatus: function(uid, cb) {
        if (typeof uid !== "string") {
            return cb(errs.ERR_INVALID_UID);
        }

        var self = this;
        var taskData = taskTracker[uid];
        if (typeof taskData === "undefined") {
            return cb(errs.ERR_TASK_DOES_NOT_EXIST);
        }
        
        Billon.SoapService.onRequestSend = self.OnRequestSend;
        Billon.SoapService.call.getTaskStatus({
            cNode: taskData.cNode ? taskData.cNode : taskData.WUI,
            params: {
                taskId: taskData.taskId
            }
        }, function(taskResponse) {

            if (typeof taskResponse.status === "string" && taskResponse.status === "FINISHED_OK") {
                var publicCodes = [];
                if (taskData && taskData.keyCodes) {
                    for (var i = 0; i < taskData.keyCodes.length; i++) {
                        publicCodes.push(taskData.keyCodes[i].Value);
                    }
                }
                taskResponse.agent = {
                    download: taskData.download,
                    keyCodes: publicCodes
                };

                if (cb) {
                    cb(taskResponse);
                }
            } else if (cb) {
                cb(taskResponse);
            }
        });
    }
};

Billon.Shop = Billon.Payment;
billon = Billon;

module.exports = Billon;
